use Link;
use searchers::Searcher;

use reqwest;
use reqwest::Client;

use serde_json;
use serde_json::Value;

use std;
use std::io::Read;

use std::any::Any;

#[derive(Debug)]
pub struct YugiohCard {
    name: String,
    card_type: String,
    text: String,
    subtype: Option<String>,
    family: Option<String>,
    atk: Option<u64>,
    def: Option<u64>,
    level: Option<u64>,
    url: String,
    image_url: String
}

impl Link for YugiohCard {
    fn label (&self) -> &str {
        &self.name
    }

    fn url (&self) -> &str {
        &self.url
    }

    fn as_any (&self) -> &Any {
        self
    }
}

pub struct YugiohSearcher {
    client: Client
}

impl YugiohSearcher {
    pub fn new () -> YugiohSearcher {
        YugiohSearcher {
            client: Client::new().unwrap()
        }
    }

    fn do_search (&self, name: &str) -> Result<String, Error> {
        let mut contents = String::new();
        let api_url = &format!("http://yugiohprices.com/api/card_data/{}", name);
        self.client.get(api_url).send()?.read_to_string(&mut contents)?;
        Result::Ok(contents)
    }
}

fn parse_entry (page: String) -> Result<YugiohCard, Error> {
    let parsed: Value = serde_json::from_str(&page)?;
    let ref parsed_entry = parsed["data"];
    if let Some(_) = parsed_entry.as_object() {
        let card_name = parsed_entry["name"].as_str().map(String::from).expect("expected name in json data");
        let card_url = format!("http://yugioh.wikia.com/wiki/{}", card_name);
        let card_image_url = format!("http://static.api3.studiobebop.net/ygo_data/card_images/{}.jpg", card_name.replace(" ", "_"));
        Result::Ok(YugiohCard {
            name: card_name,
            card_type: parsed_entry["card_type"].as_str().map(String::from).expect("expected card type in json data"),
            text: parsed_entry["text"].as_str().map(String::from).expect("expected text in json data"),
            subtype: parsed_entry["type"].as_str().map(String::from),
            family: parsed_entry["family"].as_str().map(String::from),
            atk: parsed_entry["atk"].as_u64(),
            def: parsed_entry["def"].as_u64(),
            level: parsed_entry["level"].as_u64(),
            url: card_url,
            image_url: card_image_url
        })
    } else {
        Result::Err(Error::Other(String::from("No card info found")))
    }
}

impl Searcher<YugiohCard> for YugiohSearcher {
    fn exact_search (&self, name: &str) -> Option<YugiohCard> {
        self.do_search(name).and_then(parse_entry).ok()
    }
}

#[derive(Debug)]
enum Error {
    Http(reqwest::Error),
    Io(std::io::Error),
    Json(serde_json::Error),
    Other(String)
}

impl From<reqwest::Error> for Error {
    fn from (error: reqwest::Error) -> Error {
        Error::Http(error)
    }
}

impl From<std::io::Error> for Error {
    fn from (error: std::io::Error) -> Error {
        Error::Io(error)
    }
}

impl From<serde_json::Error> for Error {
    fn from (error: serde_json::Error) -> Error {
        Error::Json(error)
    }
}
